import gixy
from gixy.core.issue import Issue


class Plugin(object):
    summary = ''
    description = ''
    help_url = ''
    severity = gixy.severity.UNSPECIFIED
    directives = []
    options = {}

    def __init__(self, config):
        self._issues = []
        self.config = config

    def add_issue(self, directive, summary=None, severity=None, description=None, reason=None, help_url=None):
        self._issues.append(Issue(self, directives=directive, summary=summary, severity=severity,
                                  description=description, reason=reason, help_url=help_url))

    def audit(self, directive):
        pass

    @property
    def issues(self):
        return self._issues

    @property
    def name(self):
        return self.__class__.__name__



import gixy
from gixy.plugins.plugin import Plugin


class valid_referers(Plugin):
    """
    Insecure example:
        valid_referers none server_names *.webvisor.com;
    """
    summary = 'Used "none" as valid referer.'
    severity = gixy.severity.HIGH
    description = 'Never trust undefined referer.'
    help_url = 'https://github.com/yandex/gixy/blob/master/docs/ru/plugins/validreferers.md'
    directives = ['valid_referers']

    def audit(self, directive):
        if 'none' in directive.args:
            self.add_issue(directive=directive)



# flake8: noqa

from configargparse import *
from six.moves import StringIO

from gixy.core.plugins_manager import PluginsManager


# used while parsing args to keep track of where they came from
_COMMAND_LINE_SOURCE_KEY = 'command_line'
_ENV_VAR_SOURCE_KEY = 'environment_variables'
_CONFIG_FILE_SOURCE_KEY = 'config_file'
_DEFAULTS_SOURCE_KEY = 'defaults'


class GixyConfigParser(DefaultConfigFileParser):
    def get_syntax_description(self):
        return ''

    def parse(self, stream):
        """Parses the keys + values from a config file."""

        items = OrderedDict()
        prefix = ''
        for i, line in enumerate(stream):
            line = line.strip()
            if not line or line[0] in ['#', ';'] or line.startswith('---'):
                continue
            if line[0] == '[':
                prefix = '%s-' % line[1:-1].replace('_', '-')
                continue

            white_space = '\\s*'
            key = '(?P<key>[^:=;#\s]+?)'
            value = white_space+'[:=\s]'+white_space+'(?P<value>.+?)'
            comment = white_space+'(?P<comment>\\s[;#].*)?'

            key_only_match = re.match('^' + key + comment + '$', line)
            if key_only_match:
                key = key_only_match.group('key')
                items[key] = 'true'
                continue

            key_value_match = re.match('^'+key+value+comment+'$', line)
            if key_value_match:
                key = key_value_match.group('key')
                value = key_value_match.group('value')

                if value.startswith('[') and value.endswith(']'):
                    # handle special case of lists
                    value = [elem.strip() for elem in value[1:-1].split(',')]

                items[prefix + key] = value
                continue

            raise ConfigFileParserException('Unexpected line %s in %s: %s' % (i,
                getattr(stream, 'name', 'stream'), line))
        return items

    def serialize(self, items):
        """Does the inverse of config parsing by taking parsed values and
        converting them back to a string representing config file contents.
        """
        r = StringIO()
        for key, value in items.items():
            if type(value) == OrderedDict:
                r.write('\n[%s]\n' % key)
                r.write(self.serialize(value))
            else:
                value, help = value
                if help:
                    r.write('; %s\n' % help)
                r.write('%s = %s\n' % (key, value))
        return r.getvalue()


class GixyHelpFormatter(HelpFormatter):
    def format_help(self):
        manager = PluginsManager()
        help_message = super(GixyHelpFormatter, self).format_help()
        if 'plugins options:' in help_message:
            # Print available blugins _only_ if we prints options for it
            plugins = '\n'.join('\t' + plugin.__name__ for plugin in manager.plugins_classes)
            help_message = '{orig}\n\navailable plugins:\n{plugins}\n'.format(orig=help_message, plugins=plugins)
        return help_message


class ArgsParser(ArgumentParser):
    def get_possible_config_keys(self, action):
        """This method decides which actions can be set in a config file and
        what their keys will be. It returns a list of 0 or more config keys that
        can be used to set the given action's value in a config file.
        """
        keys = []
        for arg in action.option_strings:
            if arg in {'--config', '--write-config', '--version'}:
                continue
            if any([arg.startswith(2*c) for c in self.prefix_chars]):
                keys += [arg[2:], arg]  # eg. for '--bla' return ['bla', '--bla']

        return keys

    def get_items_for_config_file_output(self, source_to_settings,
                                         parsed_namespace):
        """Converts the given settings back to a dictionary that can be passed
        to ConfigFormatParser.serialize(..).

        Args:
            source_to_settings: the dictionary described in parse_known_args()
            parsed_namespace: namespace object created within parse_known_args()
        Returns:
            an OrderedDict where keys are strings and values are either strings
            or lists
        """
        config_file_items = OrderedDict()
        for source, settings in source_to_settings.items():
            if source == _COMMAND_LINE_SOURCE_KEY:
                _, existing_command_line_args = settings['']
                for action in self._actions:
                    config_file_keys = self.get_possible_config_keys(action)
                    if config_file_keys and not action.is_positional_arg and \
                        already_on_command_line(existing_command_line_args,
                                                action.option_strings):
                        value = getattr(parsed_namespace, action.dest, None)
                        if value is not None:
                            if type(value) is bool:
                                value = str(value).lower()
                            if ':' in action.dest:
                                section, key = action.dest.split(':', 2)
                                key = key.replace('_', '-')
                                if section not in config_file_items:
                                    config_file_items[section] = OrderedDict()
                                config_file_items[section][key] = (value, action.help)
                            else:
                                config_file_items[config_file_keys[0]] = (value, action.help)
            elif source.startswith(_CONFIG_FILE_SOURCE_KEY):
                for key, (action, value) in settings.items():
                    if ':' in action.dest:
                        section, key = action.dest.split(':', 2)
                        key = key.replace('_', '-')
                        if section not in config_file_items:
                            config_file_items[section] = OrderedDict()
                        config_file_items[section][key] = (value, action.help)
                    else:
                        config_file_items[key] = (value, action.help)
        return config_file_items


def create_parser():
    return ArgsParser(
        description='Gixy - a Nginx configuration [sec]analyzer\n\n',
        formatter_class=GixyHelpFormatter,
        config_file_parser_class=GixyConfigParser,
        auto_env_var_prefix='GIXY_',
        add_env_var_help=False,
        default_config_files=['/etc/gixy/gixy.cfg', '~/.config/gixy/gixy.conf'],
        args_for_setting_config_path=['-c', '--config'],
        args_for_writing_out_config_file=['--write-config'],
        add_config_file_help=False
    )


import os

import gixy
from gixy.plugins.plugin import Plugin


class PluginsManager(object):

    def __init__(self, config=None):
        self.imported = False
        self.config = config
        self._plugins = []

    def import_plugins(self):
        if self.imported:
            return

        files_list = os.listdir(os.path.join(os.path.dirname(os.path.abspath(__file__)), '..', 'plugins'))
        for plugin_file in files_list:
            if not plugin_file.endswith('.py') or plugin_file.startswith('_'):
                continue
            __import__('gixy.plugins.'+os.path.splitext(plugin_file)[0], None, None, [''])

        self.imported = True

    def init_plugins(self):
        self.import_plugins()

        exclude = self.config.skips if self.config else None
        include = self.config.plugins if self.config else None
        severity = self.config.severity if self.config else None
        for plugin_cls in Plugin.__subclasses__():
            name = plugin_cls.__name__
            if include and name not in include:
                # Skip not needed plugins
                continue
            if exclude and name in exclude:
                # Skipped plugins
                continue
            if severity and not gixy.severity.is_acceptable(plugin_cls.severity, severity):
                # Skip plugin by severity level
                continue
            if self.config and self.config.has_for(name):
                options = self.config.get_for(name)
            else:
                options = plugin_cls.options
            self._plugins.append(plugin_cls(options))

    @property
    def plugins(self):
        if not self._plugins:
            self.init_plugins()
        return self._plugins

    @property
    def plugins_classes(self):
        self.import_plugins()
        return Plugin.__subclasses__()

    def get_plugins_descriptions(self):
        return map(lambda a: a.name, self.plugins)

    def audit(self, directive):
        for plugin in self.plugins:
            if plugin.directives and directive.name not in plugin.directives:
                continue
            plugin.audit(directive)

    def issues(self):
        result = []
        for plugin in self.plugins:
            if not plugin.issues:
                continue
            result.extend(plugin.issues)
        return result



import logging
import copy

from gixy.core.utils import is_indexed_name


LOG = logging.getLogger(__name__)

CONTEXTS = []


def get_context():
    return CONTEXTS[-1]


def purge_context():
    del CONTEXTS[:]


def push_context(block):
    if len(CONTEXTS):
        context = copy.deepcopy(get_context())
    else:
        context = Context()
    context.set_block(block)
    CONTEXTS.append(context)
    return context


def pop_context():
    return CONTEXTS.pop()


class Context(object):
    def __init__(self):
        self.block = None
        self.variables = {
            'index': {},
            'name': {}
        }

    def set_block(self, directive):
        self.block = directive
        return self

    def clear_index_vars(self):
        self.variables['index'] = {}
        return self

    def add_var(self, name, var):
        if is_indexed_name(name):
            var_type = 'index'
            name = int(name)
        else:
            var_type = 'name'

        self.variables[var_type][name] = var
        return self

    def get_var(self, name):
        if is_indexed_name(name):
            var_type = 'index'
            name = int(name)
        else:
            var_type = 'name'

        result = None
        try:
            result = self.variables[var_type][name]
        except KeyError:
            if var_type == 'name':
                # Only named variables can be builtins
                import gixy.core.builtin_variables as builtins

                if builtins.is_builtin(name):
                    result = builtins.builtin_var(name)

        if not result:
            LOG.info("Can't find variable '{}'".format(name))
        return result

    def __deepcopy__(self, memo):
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        result.block = copy.copy(self.block)
        result.variables = {
            'index': copy.copy(self.variables['index']),
            'name': copy.copy(self.variables['name'])
        }
        return result



import gixy
from gixy.plugins.plugin import Plugin


class add_header_redefinition(Plugin):
    """
    Insecure example:
        server {
            add_header X-Content-Type-Options nosniff;
            location / {
                add_header X-Frame-Options DENY;
            }
        }
    """
    summary = 'Nested "add_header" drops parent headers.'
    severity = gixy.severity.MEDIUM
    description = ('"add_header" replaces ALL parent headers. '
                   'See documentation: http://nginx.org/en/docs/http/ngx_http_headers_module.html#add_header')
    help_url = 'https://github.com/yandex/gixy/blob/master/docs/ru/plugins/addheaderredefinition.md'
    directives = ['server', 'location', 'if']
    options = {'headers': {'x-frame-options',
                           'x-content-type-options',
                           'x-xss-protection',
                           'content-security-policy',
                           'strict-transport-security',
                           'cache-control'}
               }

    def __init__(self, config):
        super(add_header_redefinition, self).__init__(config)
        self.interesting_headers = self.config.get('headers')

    def audit(self, directive):
        if not directive.is_block:
            # Skip all not block directives
            return

        actual_headers = get_headers(directive)
        if not actual_headers:
            return

        for parent in directive.parents:
            parent_headers = get_headers(parent)
            if not parent_headers:
                continue

            diff = (parent_headers - actual_headers) & self.interesting_headers

            if len(diff):
                self._report_issue(directive, parent, diff)

            break

    def _report_issue(self, current, parent, diff):
        directives = []
        # Add headers from parent level
        directives.extend(parent.find('add_header'))
        # Add headers from current level
        directives.extend(current.find('add_header'))
        reason = 'Parent headers "{headers}" was dropped in current level'.format(headers='", "'.join(diff))
        self.add_issue(directive=directives, reason=reason)


def get_headers(directive):
    headers = directive.find('add_header')
    if not headers:
        return set()

    return set(map(lambda d: d.header, headers))



import gixy


class Config(object):
    def __init__(self,
                 plugins=None,
                 skips=None,
                 severity=gixy.severity.UNSPECIFIED,
                 output_format=None,
                 output_file=None,
                 allow_includes=True):

        self.severity = severity
        self.output_format = output_format
        self.output_file = output_file
        self.plugins = plugins
        self.skips = skips
        self.allow_includes = allow_includes
        self.plugins_options = {}

    def set_for(self, name, options):
        self.plugins_options[name] = options

    def get_for(self, name):
        if self.has_for(name):
            return self.plugins_options[name]
        return {}

    def has_for(self, name):
        return name in self.plugins_options



import re
import logging
import gixy
from gixy.plugins.plugin import Plugin
from gixy.core.regexp import Regexp

LOG = logging.getLogger(__name__)


class origins(Plugin):
    """
    Insecure example:
        if ($http_referer !~ "^https?://([^/]+metrika.*yandex\.ru/"){
            add_header X-Frame-Options SAMEORIGIN;
        }
    """
    summary = 'Validation regex for "origin" or "referrer" matches untrusted domain.'
    severity = gixy.severity.MEDIUM
    description = 'Improve the regular expression to match only trusted referrers.'
    help_url = 'https://github.com/yandex/gixy/blob/master/docs/ru/plugins/origins.md'
    directives = ['if']
    options = {
        'domains': ['*'],
        'https_only': False
    }

    def __init__(self, config):
        super(origins, self).__init__(config)
        if self.config.get('domains') and self.config.get('domains')[0] and self.config.get('domains')[0] != '*':
            domains = '|'.join(re.escape(d) for d in self.config.get('domains'))
        else:
            domains = '[^/.]*\.[^/]{2,7}'

        scheme = 'https{http}'.format(http=('?' if not self.config.get('https_only') else ''))
        regex = r'^{scheme}://(?:[^/.]*\.){{0,10}}(?P<domain>{domains})(?::\d*)?(?:/|\?|$)'.format(
            scheme=scheme,
            domains=domains
        )
        self.valid_re = re.compile(regex)

    def audit(self, directive):
        if directive.operand not in {'~', '~*', '!~', '!~*'}:
            # Not regexp
            return

        if directive.variable not in {'$http_referer', '$http_origin'}:
            # Not interesting
            return

        invalid_referers = set()
        regexp = Regexp(directive.value, case_sensitive=(directive.operand in {'~', '!~'}))
        for value in regexp.generate('/', anchored=True):
            if value.startswith('^'):
                value = value[1:]
            else:
                value = 'http://evil.com/' + value

            if value.endswith('$'):
                value = value[:-1]
            elif not value.endswith('/'):
                value += '.evil.com'

            valid = self.valid_re.match(value)
            if not valid or valid.group('domain') == 'evil.com':
                invalid_referers.add(value)

        if invalid_referers:
            invalid_referers = '", "'.join(invalid_referers)
            name = 'origin' if directive.variable == '$http_origin' else 'referrer'
            severity = gixy.severity.HIGH if directive.variable == '$http_origin' else gixy.severity.MEDIUM
            reason = 'Regex matches "{value}" as a valid {name}.'.format(value=invalid_referers, name=name)
            self.add_issue(directive=directive, reason=reason, severity=severity)



import os
from gixy.directives.directive import Directive


DIRECTIVES = {}


def import_directives():
    files_list = os.listdir(os.path.dirname(__file__))
    for directive_file in files_list:
        if not directive_file.endswith(".py") or directive_file.startswith('_'):
            continue
        __import__('gixy.directives.' + os.path.splitext(directive_file)[0], None, None, [''])


def get_all():
    if len(DIRECTIVES):
        return DIRECTIVES

    import_directives()
    for klass in Directive.__subclasses__():
        if not klass.nginx_name:
            continue
        DIRECTIVES[klass.nginx_name] = klass

    return DIRECTIVES



import os
import sys
import logging
import copy

import gixy
from gixy.core.manager import Manager as Gixy
from gixy.formatters import get_all as formatters
from gixy.core.plugins_manager import PluginsManager
from gixy.core.config import Config
from gixy.cli.argparser import create_parser

LOG = logging.getLogger()


def _init_logger(debug=False):
    LOG.handlers = []
    log_level = logging.DEBUG if debug else logging.INFO
    logging.captureWarnings(True)

    LOG.setLevel(log_level)
    handler = logging.StreamHandler(sys.stderr)
    handler.setFormatter(logging.Formatter('[%(module)s]\t%(levelname)s\t%(message)s'))
    LOG.addHandler(handler)
    LOG.debug("logging initialized")


def _create_plugin_help(option):
    if isinstance(option, (tuple, list, set)):
        default = ','.join(list(option))
    else:
        default = str(option)

    return 'Default: {}'.format(default)


def _get_cli_parser():
    parser = create_parser()
    parser.add_argument('nginx_file', nargs='?', type=str, default='/etc/nginx/nginx.conf', metavar='nginx.conf',
                        help='Path to nginx.conf, e.g. /etc/nginx/nginx.conf')

    parser.add_argument(
        '-v', '--version', action='version',
        version='Gixy v{}'.format(gixy.version))

    parser.add_argument(
        '-l', '--level', dest='level', action='count', default=0,
        help='Report issues of a given severity level or higher (-l for LOW, -ll for MEDIUM, -lll for HIGH)')

    default_formatter = 'console' if sys.stdout.isatty() else 'text'
    available_formatters = formatters().keys()
    parser.add_argument(
        '-f', '--format', dest='output_format', choices=available_formatters, default=default_formatter,
        type=str, help='Specify output format')

    parser.add_argument(
        '-o', '--output', dest='output_file', type=str,
        help='Write report to file')

    parser.add_argument(
        '-d', '--debug', dest='debug', action='store_true', default=False,
        help='Turn on debug mode')

    parser.add_argument(
        '--tests', dest='tests', type=str,
        help='Comma-separated list of tests to run')

    parser.add_argument(
        '--skips', dest='skips', type=str,
        help='Comma-separated list of tests to skip')

    parser.add_argument(
        '--disable-includes', dest='disable_includes', action='store_true', default=False,
        help='Disable "include" directive processing')

    group = parser.add_argument_group('plugins options')
    for plugin_cls in PluginsManager().plugins_classes:
        name = plugin_cls.__name__
        if not plugin_cls.options:
            continue

        options = copy.deepcopy(plugin_cls.options)
        for opt_key, opt_val in options.items():
            option_name = '--{plugin}-{key}'.format(plugin=name, key=opt_key).replace('_', '-')
            dst_name = '{plugin}:{key}'.format(plugin=name, key=opt_key)
            opt_type = str if isinstance(opt_val, (tuple, list, set)) else type(opt_val)
            group.add_argument(
                option_name, metavar=opt_key, dest=dst_name, type=opt_type,
                help=_create_plugin_help(opt_val)
            )

    return parser


def main():
    parser = _get_cli_parser()
    args = parser.parse_args()
    _init_logger(args.debug)

    path = os.path.expanduser(args.nginx_file)
    if not os.path.isfile(path):
        sys.stderr.write('Please specify path to Nginx configuration.\n\n')
        parser.print_help()
        sys.exit(1)

    try:
        severity = gixy.severity.ALL[args.level]
    except IndexError:
        sys.stderr.write('Too high level filtering. Maximum level: -{}\n'.format('l' * (len(gixy.severity.ALL) - 1)))
        sys.exit(1)

    if args.tests:
        tests = [x.strip() for x in args.tests.split(',')]
    else:
        tests = None

    if args.skips:
        skips = [x.strip() for x in args.skips.split(',')]
    else:
        skips = None

    config = Config(
        severity=severity,
        output_format=args.output_format,
        output_file=args.output_file,
        plugins=tests,
        skips=skips,
        allow_includes=not args.disable_includes
    )

    for plugin_cls in PluginsManager().plugins_classes:
        name = plugin_cls.__name__
        options = copy.deepcopy(plugin_cls.options)
        for opt_key, opt_val in options.items():
            option_name = '{}:{}'.format(name, opt_key)
            if option_name not in args:
                continue

            val = getattr(args, option_name)
            if val is None:
                continue

            if isinstance(opt_val, tuple):
                val = tuple([x.strip() for x in val.split(',')])
            elif isinstance(opt_val, set):
                val = set([x.strip() for x in val.split(',')])
            elif isinstance(opt_val, list):
                val = [x.strip() for x in val.split(',')]
            options[opt_key] = val
        config.set_for(name, options)

    with Gixy(config=config) as yoda:
        yoda.audit(path)
        formatted = formatters()[config.output_format]().format(yoda)
        if args.output_file:
            with open(config.output_file, 'w') as f:
                f.write(formatted)
        else:
            print(formatted)

        if sum(yoda.stats.values()) > 0:
            # If something found - exit code must be 1, otherwise 0
            sys.exit(1)
        sys.exit(0)



import re
import logging

from gixy.core.regexp import Regexp
from gixy.core.context import get_context


LOG = logging.getLogger(__name__)
# See ngx_http_script_compile in http/ngx_http_script.c
EXTRACT_RE = re.compile(r'\$([1-9]|[a-z_][a-z0-9_]*|\{[a-z0-9_]+\})', re.IGNORECASE)


def compile_script(script):
    """
    Compile Nginx script to list of variables.
    Example:
        compile_script('http://$foo:$bar') ->
            [Variable('http://'), Variable($foo), Variable(':', Variable($bar).

    :param str script: Nginx scrip.
    :return Variable[]: list of variable.
    """
    depends = []
    context = get_context()
    for i, var in enumerate(EXTRACT_RE.split(str(script))):
        if i % 2:
            # Variable
            var = var.strip('{}\x20')
            var = context.get_var(var)
            if var:
                depends.append(var)
        elif var:
            # Literal
            depends.append(Variable(name=None, value=var, have_script=False))
    return depends


class Variable(object):
    def __init__(self, name, value=None, boundary=None, provider=None, have_script=True):
        """
        Gixy Nginx variable class - parse and provide helpers to work with it.

        :param str|None name: variable name.
        :param str|Regexp value: variable value..
        :param Regexp boundary: variable boundary set.
        :param Directive provider: directive that provide variable (e.g. if, location, rewrite, etc).
        :param bool have_script: may variable have nginx script or not (mostly used to indicate a string literal).
        """

        self.name = name
        self.value = value
        self.regexp = None
        self.depends = None
        self.boundary = boundary
        self.provider = provider
        if isinstance(value, Regexp):
            self.regexp = value
        elif have_script:
            self.depends = compile_script(value)

    def can_contain(self, char):
        """
        Checks if variable can contain the specified char.

        :param str char: character to test.
        :return: True if variable can contain the specified char, False otherwise.
        """

        # First of all check boundary set
        if self.boundary and not self.boundary.can_contain(char):
            return False

        # Then regexp
        if self.regexp:
            return self.regexp.can_contain(char, skip_literal=True)

        # Then dependencies
        if self.depends:
            return any(dep.can_contain(char) for dep in self.depends)

        # Otherwise user can't control value of this variable
        return False

    def can_startswith(self, char):
        """
        Checks if variable can starts with the specified char.

        :param str char: character to test.
        :return: True if variable can starts with the specified char, False otherwise.
        """

        # First of all check boundary set
        if self.boundary and not self.boundary.can_startswith(char):
            return False

        # Then regexp
        if self.regexp:
            return self.regexp.can_startswith(char)

        # Then dependencies
        if self.depends:
            return self.depends[0].can_startswith(char)

        # Otherwise user can't control value of this variable
        return False

    def must_contain(self, char):
        """
        Checks if variable MUST contain the specified char.

        :param str char: character to test.
        :return: True if variable must contain the specified char, False otherwise.
        """

        # First of all check boundary set
        if self.boundary and self.boundary.must_contain(char):
            return True

        # Then regexp
        if self.regexp:
            return self.regexp.must_contain(char)

        # Then dependencies
        if self.depends:
            return any(dep.must_contain(char) for dep in self.depends)

        # Otherwise checks literal
        return self.value and char in self.value

    def must_startswith(self, char):
        """
        Checks if variable MUST starts with the specified char.

        :param str char: character to test.
        :return: True if variable must starts with the specified char.
        """

        # First of all check boundary set
        if self.boundary and self.boundary.must_startswith(char):
            return True

        # Then regexp
        if self.regexp:
            return self.regexp.must_startswith(char)

        # Then dependencies
        if self.depends:
            return self.depends[0].must_startswith(char)

        # Otherwise checks literal
        return self.value and self.value[0] == char

    @property
    def providers(self):
        """
        Returns list of variable provides.

        :return Directive[]: providers.
        """
        result = []
        if self.provider:
            result.append(self.provider)
        if self.depends:
            for dep in self.depends:
                result += dep.providers
        return result



import gixy
from gixy.plugins.plugin import Plugin


class host_spoofing(Plugin):
    """
    Insecure example:
        proxy_set_header Host $http_host
    """
    summary = 'The proxied Host header may be spoofed.'
    severity = gixy.severity.MEDIUM
    description = 'In most cases "$host" variable are more appropriate, just use it.'
    help_url = 'https://github.com/yandex/gixy/blob/master/docs/ru/plugins/hostspoofing.md'
    directives = ['proxy_set_header']

    def audit(self, directive):
        name, value = directive.args
        if name.lower() != 'host':
            # Not a "Host" header
            return

        if value == '$http_host' or value.startswith('$arg_'):
            self.add_issue(directive=directive)



"""Very low-level nginx config parser based on pyparsing."""
import re
import logging
from cached_property import cached_property

from pyparsing import (
    Literal, Suppress, White, Word, alphanums, Forward, Group, Optional, Combine,
    Keyword, OneOrMore, ZeroOrMore, Regex, QuotedString, nestedExpr)


LOG = logging.getLogger(__name__)


class NginxQuotedString(QuotedString):
    def __init__(self, quoteChar):
        super(NginxQuotedString, self).__init__(quoteChar, escChar='\\', multiline=True)
        # Nginx parse quoted values in special manner:
        # '^https?:\/\/yandex\.ru\/\00\'\"' -> ^https?:\/\/yandex\.ru\/\00'"
        # TODO(buglloc): research and find another special characters!

        self.escCharReplacePattern = '\\\\(\'|")'


class RawParser(object):
    """
    A class that parses nginx configuration with pyparsing
    """

    def __init__(self):
        self._script = None

    def parse(self, file_path):
        """
        Returns the parsed tree.
        """
        # Temporary, dirty hack :(
        content = open(file_path).read()
        content = re.sub(r'(if\s.+)\)\)(\s*\{)?$', '\\1) )\\2', content, flags=re.MULTILINE)
        return self.script.parseString(content, parseAll=True)
        # return self.script.parseFile(file_path, parseAll=True)

    @cached_property
    def script(self):
        # constants
        left_bracket = Suppress("{")
        right_bracket = Suppress("}")
        semicolon = Suppress(";")
        space = White().suppress()
        keyword = Word(alphanums + ".+-_/")
        path = Word(alphanums + ".-_/")
        variable = Word("$_-" + alphanums)
        value_wq = Regex(r'(?:\([^\s;]*\)|\$\{\w+\}|[^\s;(){}])+')
        value_sq = NginxQuotedString(quoteChar="'")
        value_dq = NginxQuotedString(quoteChar='"')
        value = (value_dq | value_sq | value_wq)
        # modifier for location uri [ = | ~ | ~* | ^~ ]
        location_modifier = (
            Keyword("=") |
            Keyword("~*") | Keyword("~") |
            Keyword("^~"))
        # modifier for if statement
        if_modifier = Combine(Optional("!") + (
            Keyword("=") |
            Keyword("~*") | Keyword("~") |
            (Literal("-") + (Literal("f") | Literal("d") | Literal("e") | Literal("x")))))
        condition = (
            (if_modifier + Optional(space) + value) |
            (variable + Optional(space + if_modifier + Optional(space) + value))
        )

        # rules
        include = (
            Keyword("include") +
            space +
            value +
            semicolon
        )("include")

        directive = (
            keyword +
            ZeroOrMore(space + value) +
            semicolon
        )("directive")

        file_delimiter = (
            Suppress("# configuration file ") +
            path +
            Suppress(":")
        )("file_delimiter")

        comment = (
            Suppress('#') +
            Regex(r".*")
        )("comment")

        hash_value = Group(
            value +
            ZeroOrMore(space + value) +
            semicolon
        )("hash_value")

        generic_block = Forward()
        if_block = Forward()
        location_block = Forward()
        hash_block = Forward()
        unparsed_block = Forward()

        sub_block = OneOrMore(Group(if_block |
                                    location_block |
                                    hash_block |
                                    generic_block |
                                    include |
                                    directive |
                                    file_delimiter |
                                    comment |
                                    unparsed_block))

        if_block << (
            Keyword("if") +
            Suppress("(") +
            Group(condition) +
            Suppress(")") +
            Group(
                left_bracket +
                Optional(sub_block) +
                right_bracket)
        )("block")

        location_block << (
            Keyword("location") +
            Group(
                Optional(space + location_modifier) +
                Optional(space) + value) +
            Group(
                left_bracket +
                Optional(sub_block) +
                right_bracket)
        )("block")

        hash_block << (
            keyword +
            Group(OneOrMore(space + variable)) +
            Group(
                left_bracket +
                Optional(OneOrMore(hash_value)) +
                right_bracket)
        )("block")

        generic_block << (
            keyword +
            Group(ZeroOrMore(space + variable)) +
            Group(
                left_bracket +
                Optional(sub_block) +
                right_bracket)
        )("block")

        unparsed_block << (
            keyword +
            Group(ZeroOrMore(space + variable)) +
            nestedExpr(opener="{", closer="}")
        )("unparsed_block")

        return sub_block



import six
import logging
import re
import random
import itertools
from cached_property import cached_property

import gixy.core.sre_parse.sre_parse as sre_parse

LOG = logging.getLogger(__name__)


def _build_reverse_list(original):
    result = []
    for c in range(1, 126):
        c = six.unichr(c)
        if c not in original:
            result.append(c)
    return frozenset(result)


FIX_NAMED_GROUPS_RE = re.compile(r"(?<!\\)\(\?(?:<|')(\w+)(?:>|')")

CATEGORIES = {
    # TODO(buglloc): unicode?
    sre_parse.CATEGORY_SPACE: sre_parse.WHITESPACE,
    sre_parse.CATEGORY_NOT_SPACE: _build_reverse_list(sre_parse.WHITESPACE),
    sre_parse.CATEGORY_DIGIT: sre_parse.DIGITS,
    sre_parse.CATEGORY_NOT_DIGIT: _build_reverse_list(sre_parse.DIGITS),
    sre_parse.CATEGORY_WORD: frozenset('abcdefghijklmnopqrstuvwxyz'
                                       'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
                                       '0123456789_'),
    sre_parse.CATEGORY_NOT_WORD: _build_reverse_list(frozenset('abcdefghijklmnopqrstuvwxyz'
                                                               'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
                                                               '0123456789_')),
    sre_parse.CATEGORY_LINEBREAK: frozenset('\n'),
    sre_parse.CATEGORY_NOT_LINEBREAK: _build_reverse_list(frozenset('\n')),
    'ANY': [six.unichr(x) for x in range(1, 127) if x != 10]
}

CATEGORIES_NAMES = {
    sre_parse.CATEGORY_DIGIT: r'\d',
    sre_parse.CATEGORY_NOT_DIGIT: r'\D',
    sre_parse.CATEGORY_SPACE: r'\s',
    sre_parse.CATEGORY_NOT_SPACE: r'\S',
    sre_parse.CATEGORY_WORD: r'\w',
    sre_parse.CATEGORY_NOT_WORD: r'\W',
}


def extract_groups(parsed, top=True):
    result = {}
    if top:
        result[0] = parsed
    for token in parsed:
        if not token:
            # Skip empty tokens
            pass
        elif token[0] == sre_parse.SUBPATTERN:
            if isinstance(token[1][0], int):
                # Captured group index can't be a string. E.g. for pattern "(?:la)" group name is "None"
                result[token[1][0]] = token[1][1]
            result.update(extract_groups(token[1][1], False))
        elif token[0] == sre_parse.MIN_REPEAT:
            result.update(extract_groups(token[1][2], False))
        elif token[0] == sre_parse.MAX_REPEAT:
            result.update(extract_groups(token[1][2], False))
        elif token[0] == sre_parse.BRANCH:
            result.update(extract_groups(token[1][1], False))
        elif token[0] == sre_parse.SUBPATTERN:
            result.update(extract_groups(token[1][1], False))
        elif token[0] == sre_parse.IN:
            result.update(extract_groups(token[1], False))
        elif isinstance(token, sre_parse.SubPattern):
            result.update(extract_groups(token, False))
    return result


def _gen_combinator(variants, _merge=True):
    if not hasattr(variants, '__iter__'):
        return [variants] if variants is not None else []

    res = []
    need_product = False
    for var in variants:
        if isinstance(var, list):
            sol = _gen_combinator(var, _merge=False)
            res.append(sol)
            need_product = True
        elif var is not None:
            res.append(var)

    if need_product:
        producted = itertools.product(*res)
        if _merge:
            # TODO(buglloc): ??!
            return list(six.moves.map(_merge_variants, producted))
        return producted
    elif _merge:
        return list(six.moves.map(_merge_variants, [res]))
    return res


def _merge_variants(variants):
    result = []
    for var in variants:
        if isinstance(var, tuple):
            result.append(_merge_variants(var))
        else:
            result.append(var)
    return ''.join(result)


class Token(object):
    type = None

    def __init__(self, token, parent, regexp):
        self.token = token
        self.childs = None
        self.parent = parent
        self.regexp = regexp
        self._parse()

    def parse(self):
        pass

    def _parse(self):
        pass

    def _parse_childs(self, childs):
        self.childs = parse(childs, self, regexp=self.regexp)

    def _get_group(self, gid):
        return self.regexp.group(gid)

    def _reg_group(self, gid):
        self.regexp.reg_group(gid, self)

    def can_contain(self, char, skip_literal=True):
        raise NotImplementedError('can_contain must be implemented')

    def can_startswith(self, char, strict=False):
        return self.can_contain(char, skip_literal=False)

    def must_contain(self, char):
        raise NotImplementedError('must_contain must be implemented')

    def must_startswith(self, char, strict=False):
        return self.must_contain(char)

    def generate(self, context):
        raise NotImplementedError('generate must be implemented')

    def __str__(self):
        raise NotImplementedError('__str__ must be implemented')


class AnyToken(Token):
    type = sre_parse.ANY

    def can_contain(self, char, skip_literal=True):
        return char in CATEGORIES['ANY']

    def must_contain(self, char, skip_literal=True):
        # Char may not be present in ANY token
        return False

    def generate(self, context):
        if context.char in CATEGORIES['ANY']:
            return context.char
        return 'a'

    def __str__(self):
        return '.'


class LiteralToken(Token):
    type = sre_parse.LITERAL

    def _parse(self):
        self.char = six.unichr(self.token[1])

    def can_contain(self, char, skip_literal=True):
        if skip_literal:
            return False
        return self.char == char

    def must_contain(self, char, skip_literal=True):
        return self.char == char

    def generate(self, context):
        return self.char

    def __str__(self):
        return re.escape(self.char)


class NotLiteralToken(Token):
    type = sre_parse.NOT_LITERAL

    def _parse(self):
        self.char = six.unichr(self.token[1])
        self.gen_char_list = list(_build_reverse_list(frozenset(self.char)))

    def can_contain(self, char, skip_literal=True):
        return self.char != char

    def must_contain(self, char):
        # Any char MAY not be present in NotLiteral, e.g.: "a" not present in "[^b]"
        return False

    def generate(self, context):
        if self.can_contain(context.char):
            return context.char

        return random.choice(self.gen_char_list)

    def __str__(self):
        return '[^{char}]'.format(char=self.char)


class RangeToken(Token):
    type = sre_parse.RANGE

    def _parse(self):
        self.left_code = self.token[1][0]
        self.right_code = self.token[1][1]
        self.left = six.unichr(self.left_code)
        self.right = six.unichr(self.right_code)

    def can_contain(self, char, skip_literal=True):
        return self.left <= char <= self.right

    def must_contain(self, char, skip_literal=True):
        return self.left == char == self.right

    def generate(self, context):
        if self.can_contain(context.char):
            return context.char

        return six.unichr(random.randint(self.token[1][0], self.token[1][1]))

    def __str__(self):
        return '{left}-{right}'.format(left=self.left, right=self.right)


class CategoryToken(Token):
    type = sre_parse.CATEGORY

    def _parse(self):
        self.char_list = CATEGORIES.get(self.token[1], [''])

    def can_contain(self, char, skip_literal=True):
        return char in self.char_list

    def must_contain(self, char, skip_literal=True):
        return frozenset([char]) == self.char_list

    def generate(self, context):
        if self.can_contain(context.char):
            return context.char

        for c in self.char_list:
            return c

    def __str__(self):
        return CATEGORIES_NAMES.get(self.token[1], '\\C')


class MinRepeatToken(Token):
    type = sre_parse.MIN_REPEAT

    def _parse(self):
        self._parse_childs(self.token[1][2])
        self.min = self.token[1][0]
        self.max = self.token[1][1]

    def can_contain(self, char, skip_literal=True):
        if self.max == 0:
            # [a-z]{0}
            return False
        for child in self.childs:
            if child.can_contain(char, skip_literal=skip_literal):
                return True
        return False

    def must_contain(self, char):
        if self.max == 0:
            # [a-z]{0}
            return False
        if self.min == 0:
            # [a-z]*?
            return False
        for child in self.childs:
            if child.must_contain(char):
                return True
        return False

    def can_startswith(self, char, strict=False):
        if self.max == 0:
            # [a-z]{0}
            if self.childs[0].can_startswith(char, strict):
                return False
            return None
        return self.childs[0].can_startswith(char, strict)

    def must_startswith(self, char, strict=False):
        if self.min == 0:
            # [a-z]*?
            return None
        if self.max == 0:
            # [a-z]{0}
            return None
        return self.childs[0].must_startswith(char, strict=strict)

    def generate(self, context):
        res = []
        if self.min == 0:
            # [a-z]*
            res.append('')
        if self.max == 0:
            # [a-z]{0}
            return res

        for child in self.childs:
            res.extend(child.generate(context))

        result = []
        repeat = self.max if self.max <= context.max_repeat else context.max_repeat
        for val in _gen_combinator([res]):
            result.append(val * repeat)
        return result

    def __str__(self):
        childs = ''.join(str(x) for x in self.childs)
        if self.min == self.max:
            return '{childs}{{{count}}}?'.format(childs=childs, count=self.min)
        if self.min == 0 and self.max == 1:
            return '{childs}?'.format(childs=childs)
        if self.min == 0 and self.max == sre_parse.MAXREPEAT:
            return '{childs}*?'.format(childs=childs)
        if self.min == 1 and self.max == sre_parse.MAXREPEAT:
            return '{childs}+?'.format(childs=childs)
        return '{childs}{{{min},{max}}}?'.format(childs=childs, min=self.min, max=self.max)


class MaxRepeatToken(Token):
    type = sre_parse.MAX_REPEAT

    def _parse(self):
        self._parse_childs(self.token[1][2])
        self.min = self.token[1][0]
        self.max = self.token[1][1]

    def can_contain(self, char, skip_literal=True):
        if self.max == 0:
            # [a-z]{0}
            return False
        for child in self.childs:
            if child.can_contain(char, skip_literal=skip_literal):
                return True
        return False

    def must_contain(self, char):
        if self.max == 0:
            # [a-z]{0}
            return False
        if self.min == 0:
            # [a-z]?
            return False
        for child in self.childs:
            if child.must_contain(char):
                return True
        return False

    def can_startswith(self, char, strict=False):
        if self.max == 0:
            # [a-z]{0}
            if self.childs[0].can_startswith(char, strict):
                return False
            return None
        return self.childs[0].can_startswith(char, strict)

    def must_startswith(self, char, strict=False):
        if self.min == 0:
            # [a-z]*
            return None
        if self.max == 0:
            # [a-z]{0}
            return None
        return self.childs[0].must_startswith(char, strict=strict)

    def generate(self, context):
        res = []
        if self.min == 0:
            # [a-z]*
            res.append('')
        if self.max == 0:
            # [a-z]{0}
            return res

        for child in self.childs:
            res.extend(child.generate(context))

        result = []
        repeat = self.max if self.max <= context.max_repeat else context.max_repeat
        for val in _gen_combinator([res]):
            result.append(val * repeat)
        return result

    def __str__(self):
        childs = ''.join(str(x) for x in self.childs)
        if self.min == self.max:
            return '{childs}{{{count}}}'.format(childs=childs, count=self.min)
        if self.min == 0 and self.max == 1:
            return '{childs}?'.format(childs=childs)
        if self.min == 0 and self.max == sre_parse.MAXREPEAT:
            return '{childs}*'.format(childs=childs)
        if self.min == 1 and self.max == sre_parse.MAXREPEAT:
            return '{childs}+'.format(childs=childs)
        return '{childs}{{{min},{max}}}'.format(childs=childs, min=self.min, max=self.max)


class BranchToken(Token):
    type = sre_parse.BRANCH

    def _parse(self):
        self.childs = []
        for token in self.token[1][1]:
            if not token:
                self.childs.append(EmptyToken(token=token, parent=self.parent, regexp=self.regexp))
            elif isinstance(token, sre_parse.SubPattern):
                self.childs.append(InternalSubpatternToken(token=token, parent=self.parent, regexp=self.regexp))
            else:
                raise RuntimeError('Unexpected token {} in branch'.format(token))

    def can_contain(self, char, skip_literal=True):
        for child in self.childs:
            if child.can_contain(char, skip_literal=skip_literal):
                return True
        return False

    def must_contain(self, char):
        return all(child.must_contain(char) for child in self.childs)

    def can_startswith(self, char, strict=False):
        return any(x.can_startswith(char, strict) for x in self.childs)

    def must_startswith(self, char, strict=False):
        return all(x.must_startswith(char, strict) for x in self.childs)

    def generate(self, context):
        res = []
        for child in self.childs:
            values = child.generate(context)
            if isinstance(values, list):
                res.extend(child.generate(context))
            else:
                res.append(values)

        return res

    def __str__(self):
        return '(?:{})'.format('|'.join(str(x) for x in self.childs))


class SubpatternToken(Token):
    type = sre_parse.SUBPATTERN

    def _parse(self):
        self._parse_childs(self.token[1][1])
        self.group = self.token[1][0]
        if isinstance(self.group, int):
            # Captured group index can't be a string. E.g. for pattern "(?:la)" group name is "None"
            self._reg_group(self.group)

    def can_contain(self, char, skip_literal=True):
        for child in self.childs:
            if child.can_contain(char, skip_literal=skip_literal):
                return True
        return False

    def must_contain(self, char):
        for child in self.childs:
            if child.must_contain(char):
                return True
        return False

    def can_startswith(self, char, strict=False):
        if isinstance(self.childs[0], AtToken):
            if len(self.childs) > 1:
                for child in self.childs[1:]:
                    can = child.can_startswith(char, strict)
                    if can is None:
                        continue
                    return can
            return False
        elif not strict and not isinstance(self.childs[0], (SubpatternToken, InternalSubpatternToken)):
            # Not strict regexp w/o ^ can starts with any character
            return char in CATEGORIES['ANY']

        for child in self.childs:
            can = child.can_startswith(char, strict)
            if can is None:
                continue
            return can
        return None

    def must_startswith(self, char, strict=False):
        if isinstance(self.childs[0], AtToken):
            if len(self.childs) > 1:
                for child in self.childs[1:]:
                    must = child.must_startswith(char, strict=True)
                    if must is None:
                        continue
                    return must
            return False
        elif not strict and not isinstance(self.childs[0], (SubpatternToken, InternalSubpatternToken)):
            # Not strict regexp w/o ^ MAY NOT starts with any character
            return False

        for child in self.childs:
            must = child.must_startswith(char, strict=strict)
            if must is None:
                continue
            return must
        return None

    def generate(self, context):
        res = []
        for child in self.childs:
            res.append(child.generate(context))

        return _gen_combinator(res)

    def __str__(self):
        childs = ''.join(str(x) for x in self.childs)
        if self.group is None:
            return '(?:{childs})'.format(childs=childs)
        return '({childs})'.format(childs=childs)


class InternalSubpatternToken(Token):
    type = sre_parse.SUBPATTERN

    def _parse(self):
        self._parse_childs(self.token)
        self.group = None

    def can_contain(self, char, skip_literal=True):
        for child in self.childs:
            if child.can_contain(char, skip_literal=skip_literal):
                return True
        return False

    def must_contain(self, char):
        for child in self.childs:
            if child.must_contain(char):
                return True
        return False

    def can_startswith(self, char, strict=False):
        if isinstance(self.childs[0], AtToken):
            if len(self.childs) > 1:
                for child in self.childs[1:]:
                    can = child.can_startswith(char, strict)
                    if can is None:
                        continue
                    return can
            return False
        elif not strict and not isinstance(self.childs[0], (SubpatternToken, InternalSubpatternToken)):
            # Not strict regexp w/o ^ can starts with any character
            return char in CATEGORIES['ANY']

        for child in self.childs:
            can = child.can_startswith(char, strict)
            if can is None:
                continue
            return can
        return None

    def must_startswith(self, char, strict=False):
        if isinstance(self.childs[0], AtToken):
            if len(self.childs) > 1:
                for child in self.childs[1:]:
                    must = child.must_startswith(char, strict=True)
                    if must is None:
                        continue
                    return must
            return False
        elif not strict and not isinstance(self.childs[0], (SubpatternToken, InternalSubpatternToken)):
            # Not strict regexp w/o ^ MAY NOT starts with any character
            return False

        for child in self.childs:
            must = child.must_startswith(char, strict=strict)
            if must is None:
                continue
            return must
        return None

    def generate(self, context):
        res = []
        for child in self.childs:
            res.append(child.generate(context))

        return _gen_combinator(res)

    def __str__(self):
        return ''.join(str(x) for x in self.childs)


class InToken(Token):
    type = sre_parse.IN

    def _parse(self):
        self.childs = parse(self.token[1], self)

    def can_contain(self, char, skip_literal=True):
        can = False
        negative = False
        for child in self.childs:
            if isinstance(child, NegateToken):
                negative = True
            else:
                can = child.can_contain(char, skip_literal=False)

            if can:
                break
        if can and not negative:
            # a in [a-z]
            return True
        if not can and negative:
            # a in [^b-z]
            return True
        return False

    def must_contain(self, char):
        # Any character MAY not be present in IN
        return False

    def _generate_positive(self, context):
        result = []
        for child in self.childs:
            if isinstance(child, (NegateToken, EmptyToken)):
                pass
            else:
                result.append(child.generate(context=context))
        return result

    def _generate_negative(self, context):
        blacklisted = set()
        # TODO(buglloc): move chars list into the tokens?
        for child in self.childs:
            if isinstance(child, (NegateToken, EmptyToken)):
                pass
            elif isinstance(child, LiteralToken):
                blacklisted.add(child.char)
            elif isinstance(child, RangeToken):
                blacklisted.update(six.unichr(c) for c in six.moves.range(child.left_code, child.right_code + 1))
            elif isinstance(child, CategoryToken):
                blacklisted.update(child.char_list)
            else:
                LOG.info('Unexpected child "{!r}"'.format(child))

        for char in _build_reverse_list(set()):
            if char not in blacklisted:
                return char

    def generate(self, context):
        if self.can_contain(context.char, skip_literal=False):
            return context.char

        is_negative = self.childs and isinstance(self.childs[0], NegateToken)
        if is_negative:
            # [^a-z]
            return self._generate_negative(context)
        # [a-z]
        return self._generate_positive(context)

    def __str__(self):
        return '[{childs}]'.format(childs=''.join(str(x) for x in self.childs))


class AtToken(Token):
    type = sre_parse.AT

    def _parse(self):
        self.begin = self.token[1] == sre_parse.AT_BEGINNING
        self.end = self.token[1] == sre_parse.AT_END

    def can_contain(self, char, skip_literal=True):
        return False

    def must_contain(self, char):
        return False

    def generate(self, context):
        if context.anchored:
            if self.begin:
                return '^'
            if self.end:
                return '$'
        return None

    def __str__(self):
        if self.begin:
            return '^'
        if self.end:
            return '$'
        LOG.warn('unexpected AT token: %s', self.token)


class NegateToken(Token):
    type = sre_parse.NEGATE

    def can_contain(self, char, skip_literal=True):
        return False

    def must_contain(self, char):
        return False

    def can_startswith(self, char, strict=False):
        return None

    def must_startswith(self, char, strict=False):
        return None

    def generate(self, context):
        return None

    def __str__(self):
        return '^'


class GroupRefToken(Token):
    type = sre_parse.GROUPREF

    def _parse(self):
        self.id = self.token[1]
        self.group = self._get_group(self.id)

    def can_contain(self, char, skip_literal=True):
        return self.group.can_contain(char, skip_literal=skip_literal)

    def must_contain(self, char):
        return self.group.must_contain(char)

    def can_startswith(self, char, strict=False):
        return self.group.can_startswith(char, strict=strict)

    def must_startswith(self, char, strict=False):
        return self.group.must_startswith(char, strict=strict)

    def generate(self, context):
        return self.group.generate(context)

    def __str__(self):
        return '\\\\{}'.format(self.id)


class AssertToken(Token):
    type = sre_parse.ASSERT

    def can_contain(self, char, skip_literal=True):
        # TODO(buglloc): Do it!
        return False

    def must_contain(self, char):
        # TODO(buglloc): Do it!
        return False

    def can_startswith(self, char, strict=False):
        return None

    def must_startswith(self, char, strict=False):
        return None


class AssertNotToken(Token):
    type = sre_parse.ASSERT_NOT

    def can_contain(self, char, skip_literal=True):
        # TODO(buglloc): Do it!
        return False

    def must_contain(self, char):
        # TODO(buglloc): Do it!
        return False

    def can_startswith(self, char, strict=False):
        return None

    def must_startswith(self, char, strict=False):
        return None


class EmptyToken(Token):
    type = None

    def can_contain(self, char, skip_literal=True):
        return False

    def must_contain(self, char):
        # TODO(buglloc): Do it!
        return False

    def can_startswith(self, char, strict=False):
        return None

    def must_startswith(self, char, strict=False):
        return None

    def generate(self, context):
        return ''

    def __str__(self):
        return ''


def parse(sre_obj, parent=None, regexp=None):
    result = []
    for token in sre_obj:
        if not token:
            result.append(EmptyToken(token=token, parent=parent, regexp=regexp))
        elif token[0] == sre_parse.ANY:
            result.append(AnyToken(token=token, parent=parent, regexp=regexp))
        elif token[0] == sre_parse.LITERAL:
            result.append(LiteralToken(token=token, parent=parent, regexp=regexp))
        elif token[0] == sre_parse.NOT_LITERAL:
            result.append(NotLiteralToken(token=token, parent=parent, regexp=regexp))
        elif token[0] == sre_parse.RANGE:
            result.append(RangeToken(token=token, parent=parent, regexp=regexp))
        elif token[0] == sre_parse.CATEGORY:
            result.append(CategoryToken(token=token, parent=parent, regexp=regexp))
        elif token[0] == sre_parse.MIN_REPEAT:
            result.append(MinRepeatToken(token=token, parent=parent, regexp=regexp))
        elif token[0] == sre_parse.MAX_REPEAT:
            result.append(MaxRepeatToken(token=token, parent=parent, regexp=regexp))
        elif token[0] == sre_parse.BRANCH:
            result.append(BranchToken(token=token, parent=parent, regexp=regexp))
        elif token[0] == sre_parse.SUBPATTERN:
            result.append(SubpatternToken(token=token, parent=parent, regexp=regexp))
        elif token[0] == sre_parse.IN:
            result.append(InToken(token=token, parent=parent, regexp=regexp))
        elif token[0] == sre_parse.NEGATE:
            result.append(NegateToken(token=token, parent=parent, regexp=regexp))
        elif token[0] == sre_parse.AT:
            result.append(AtToken(token=token, parent=parent, regexp=regexp))
        elif token[0] == sre_parse.GROUPREF:
            result.append(GroupRefToken(token=token, parent=parent, regexp=regexp))
        elif token[0] == sre_parse.ASSERT:
            pass  # TODO(buglloc): Do it!
        elif token[0] == sre_parse.ASSERT_NOT:
            pass  # TODO(buglloc): Do it!
        else:
            LOG.info('Unexpected token "{}"'.format(token[0]))

    return result


class GenerationContext(object):
    def __init__(self, char, max_repeat=5, strict=False, anchored=True):
        self.char = char
        self.max_repeat = max_repeat
        self.strict = strict
        self.anchored = anchored


class Regexp(object):
    def __init__(self, source, strict=False, case_sensitive=True, _root=None, _parsed=None):
        """
        Gixy Regexp class, parse and provide helpers to work with it.

        :param str source: regexp, e.g. ^foo$.
        :param bool strict: anchored or not.
        :param bool case_sensitive: case sensitive or not.
        """

        self.source = source
        self.strict = strict
        self.case_sensitive = case_sensitive
        self._root = _root
        self._parsed = _parsed
        self._groups = {}

    def can_startswith(self, char):
        """
        Checks if regex can starts with the specified char.
        Example:
          Regexp('[a-z][0-9]').can_startswith('s') -> True
          Regexp('[a-z][0-9]').can_startswith('0') -> True
          Regexp('^[a-z][0-9]').can_startswith('0') -> False
          Regexp('[a-z][0-9]', strict=True).can_startswith('0') -> False

        :param str char: character to test.
        :return bool: True if regex can starts with the specified char, False otherwise.
        """

        return self.root.can_startswith(
            char=char if self.case_sensitive else char.lower(),
            strict=self.strict
        )

    def can_contain(self, char, skip_literal=True):
        """
        Checks if regex can contain the specified char.
        Example:
          Regexp('[a-z][0-9]').can_contain('s') -> True
          Regexp('[a-z][0-9]').can_contain('0') -> True
          Regexp('[a-z][0-9]').can_contain('/') -> False
          Regexp('[a-z][0-9]/').can_contain('/') -> False
          Regexp('[a-z][0-9]/').can_contain('/', skip_literal=False) -> True

        :param str char: character to test.
        :param bool skip_literal: skip literal tokens.
        :return bool: True if regex can contain the specified char, False otherwise.
        """

        return self.root.can_contain(
            char=char if self.case_sensitive else char.lower(),
            skip_literal=skip_literal
        )

    def must_startswith(self, char):
        """
        Checks if regex MUST starts with the specified char.
        Example:
          Regexp('[a-z][0-9]').must_startswith('s') -> False
          Regexp('s[a-z]').must_startswith('s') -> False
          Regexp('^s[a-z]').must_startswith('s') -> True
          Regexp('s[a-z]', strict=True).must_startswith('s') -> True

        :param str char: character to test.
        :return bool: True if regex must starts with the specified char, False otherwise.
        """

        return self.root.must_startswith(
            char=char if self.case_sensitive else char.lower(),
            strict=self.strict
        )

    def must_contain(self, char):
        """
        Checks if regex MUST contain the specified char.
        Example:
          Regexp('[a-z][0-9]').must_contain('s') -> False
          Regexp('[a-z][0-9]s').must_contain('s') -> True

        :param str char: character to test.
        :return bool: True if regex MUST contain the specified char, False otherwise.
        """

        return self.root.must_contain(
            char=char if self.case_sensitive else char.lower()
        )

    def generate(self, char, anchored=False, max_repeat=5):
        """
        Generate values that match regex.
        Example:
          Regexp('.a?').generate('s') -> ['s', 'sa']
          Regexp('(?:^http|https)://.').generate('s') -> ['http://s', 'https://s']
          Regexp('(?:^http|https)://.').generate('s', anchored=True) -> ['^http://s', 'https://s']


        :param str char: "dangerous" character, generator try to place it wherever possible.
        :param bool anchored: place anchors in generated values.
        :param int max_repeat: maximum count of repeated group (e.g. "a+" provides "aaaaa").
        :return list of str: True if regex can contain the specified char, False otherwise.
        """

        context = GenerationContext(char, anchored=anchored, max_repeat=max_repeat)
        for val in self.root.generate(context=context):
            if anchored and self.strict and not val.startswith('^'):
                yield '^' + val
            else:
                yield val

    def group(self, name):
        """
        Returns group by specified name.

        :param name: name of the group.
        :return Regexp: Regexp object for this group.
        """

        if name in self.groups:
            return self.groups[name]
        return Regexp('')

    def reg_group(self, gid, token):
        self._groups[gid] = token

    def get_group(self, gid):
        return self._groups[gid]

    @cached_property
    def groups(self):
        # self.root.parse()
        result = {}
        # for name, token in self._groups.items():
        #     result[name] = Regexp(str(self), root=token, strict=True, case_sensitive=self.case_sensitive)
        for name, parsed in extract_groups(self.parsed).items():
            result[name] = Regexp('compiled', _parsed=parsed, strict=True, case_sensitive=self.case_sensitive)
        for name, group in self.parsed.pattern.groupdict.items():
            result[name] = result[group]
        return result

    @property
    def root(self):
        if self._root:
            return self._root

        self._root = InternalSubpatternToken(self.parsed, parent=None, regexp=self)
        self._groups[0] = self._root
        return self._root

    @property
    def parsed(self):
        # TODO(buglloc): Ugly hack!
        if self._parsed:
            return self._parsed

        self._parsed = sre_parse.parse(FIX_NAMED_GROUPS_RE.sub('(?P<\\1>', self.source))
        return self._parsed

    def __str__(self):
        return str(self.root)



class Issue(object):

    def __init__(self, plugin, summary=None, description=None,
                 severity=None, reason=None, help_url=None, directives=None):
        self.plugin = plugin
        self.summary = summary
        self.description = description
        self.severity = severity
        self.reason = reason
        self.help_url = help_url
        if not directives:
            self.directives = []
        elif not hasattr(directives, '__iter__'):
            self.directives = [directives]
        else:
            self.directives = directives



from gixy.core.variable import Variable
from gixy.core.regexp import Regexp


def get_overrides():
    result = {}
    for klass in Directive.__subclasses__():
        if not klass.nginx_name:
            continue

        if not klass.__name__.endswith('Directive'):
            continue

        result[klass.nginx_name] = klass
    return result


class Directive(object):
    nginx_name = None
    is_block = False
    provide_variables = False

    def __init__(self, name, args, raw=None):
        self.name = name
        self.parent = None
        self.args = args
        self._raw = raw

    def set_parent(self, parent):
        self.parent = parent

    @property
    def parents(self):
        parent = self.parent
        while parent:
            yield parent
            parent = parent.parent

    @property
    def variables(self):
        raise NotImplementedError()

    def __str__(self):
        return '{} {};'.format(self.name, ' '.join(self.args))


class AddHeaderDirective(Directive):
    nginx_name = 'add_header'

    def __init__(self, name, args):
        super(AddHeaderDirective, self).__init__(name, args)
        self.header = args[0].lower()
        self.value = args[1]
        self.always = False
        if len(args) > 2 and args[2] == 'always':
            self.always = True


class SetDirective(Directive):
    nginx_name = 'set'
    provide_variables = True

    def __init__(self, name, args):
        super(SetDirective, self).__init__(name, args)
        self.variable = args[0].strip('$')
        self.value = args[1]

    @property
    def variables(self):
        return [Variable(name=self.variable, value=self.value, provider=self)]


class SetByLuaDirective(Directive):
    nginx_name = 'set_by_lua'
    provide_variables = True

    def __init__(self, name, args):
        super(SetByLuaDirective, self).__init__(name, args)
        self.variable = args[0].strip('$')
        self.value = args[1]

    @property
    def variables(self):
        return [Variable(name=self.variable, provider=self, have_script=False)]


class RewriteDirective(Directive):
    nginx_name = 'rewrite'
    provide_variables = True
    boundary = Regexp('[^\s\r\n]')

    def __init__(self, name, args):
        super(RewriteDirective, self).__init__(name, args)
        self.pattern = args[0]
        self.replace = args[1]
        self.flag = None
        if len(args) > 2:
            self.flag = args[2]

    @property
    def variables(self):
        regexp = Regexp(self.pattern, case_sensitive=True)
        result = []
        for name, group in regexp.groups.items():
            result.append(Variable(name=name, value=group, boundary=self.boundary, provider=self))
        return result


class RootDirective(Directive):
    nginx_name = 'root'
    provide_variables = True

    def __init__(self, name, args):
        super(RootDirective, self).__init__(name, args)
        self.path = args[0]

    @property
    def variables(self):
        return [Variable(name='document_root', value=self.path, provider=self)]



from gixy.core.regexp import Regexp
from gixy.core.variable import Variable


BUILTIN_VARIABLES = {
    # http://nginx.org/en/docs/http/ngx_http_core_module.html#var_uri
    'uri': '/[^\x20\t]*',
    # http://nginx.org/en/docs/http/ngx_http_core_module.html#var_document_uri
    'document_uri': '/[^\x20\t]*',
    # http://nginx.org/en/docs/http/ngx_http_core_module.html#var_arg_
    'arg_': '[^\s&]+',
    # http://nginx.org/en/docs/http/ngx_http_core_module.html#var_args
    'args': '[^\s]+',
    # http://nginx.org/en/docs/http/ngx_http_core_module.html#var_query_string
    'query_string': '[^\s]+',
    # http://nginx.org/en/docs/http/ngx_http_core_module.html#var_request_uri
    'request_uri': '/[^\s]*',
    # http://nginx.org/en/docs/http/ngx_http_core_module.html#var_http_
    'http_': '[\x21-\x7e]',

    # http://nginx.org/en/docs/http/ngx_http_upstream_module.html#var_upstream_http_
    'upstream_http_': '',
    # http://nginx.org/en/docs/http/ngx_http_upstream_module.html#var_upstream_cookie_
    'upstream_cookie_': '',
    # http://nginx.org/en/docs/http/ngx_http_proxy_module.html#var_proxy_add_x_forwarded_for
    'proxy_add_x_forwarded_for': '',
    # http://nginx.org/en/docs/http/ngx_http_proxy_module.html#var_proxy_host
    'proxy_host': '',
    # http://nginx.org/en/docs/http/ngx_http_proxy_module.html#var_proxy_port
    'proxy_port': '',
    # http://nginx.org/en/docs/http/ngx_http_core_module.html#var_proxy_protocol_addr
    # http://nginx.org/en/docs/stream/ngx_stream_core_module.html#var_proxy_protocol_addr
    # http://nginx.org/en/docs/http/ngx_http_core_module.html#var_proxy_protocol_port
    # http://nginx.org/en/docs/stream/ngx_stream_core_module.html#var_proxy_protocol_port
    'proxy_protocol_port': '',
    # http://nginx.org/en/docs/http/ngx_http_fastcgi_module.html#var_fastcgi_path_info
    'fastcgi_path_info': '',
    # http://nginx.org/en/docs/http/ngx_http_fastcgi_module.html#var_fastcgi_script_name
    'fastcgi_script_name': '',
    # http://nginx.org/en/docs/http/ngx_http_core_module.html#var_content_type
    'content_type': '',
    # http://nginx.org/en/docs/http/ngx_http_core_module.html#var_cookie_
    'cookie_': '',
    # http://nginx.org/en/docs/http/ngx_http_core_module.html#var_host
    'host': '',
    # http://nginx.org/en/docs/http/ngx_http_core_module.html#var_hostname
    # http://nginx.org/en/docs/stream/ngx_stream_core_module.html#var_hostname
    'hostname': '',
    # http://nginx.org/en/docs/http/ngx_http_core_module.html#var_limit_rate
    'limit_rate': '',
    # http://nginx.org/en/docs/http/ngx_http_memcached_module.html#var_memcached_key
    'memcached_key': '',
    # http://nginx.org/en/docs/http/ngx_http_core_module.html#var_realpath_root
    'realpath_root': '',
    # http://nginx.org/en/docs/http/ngx_http_core_module.html#var_remote_user
    'remote_user': '',
    # http://nginx.org/en/docs/http/ngx_http_core_module.html#var_request
    'request': '',
    # http://nginx.org/en/docs/http/ngx_http_core_module.html#var_request_body
    'request_body': '',
    # http://nginx.org/en/docs/http/ngx_http_core_module.html#var_request_completion
    'request_completion': '',
    # http://nginx.org/en/docs/http/ngx_http_core_module.html#var_request_filename
    'request_filename': '',
    # http://nginx.org/en/docs/http/ngx_http_core_module.html#var_request_id
    'request_id': '',
    # http://nginx.org/en/docs/http/ngx_http_slice_module.html#var_slice_range
    'slice_range': '',
    # http://nginx.org/en/docs/http/ngx_http_secure_link_module.html#var_secure_link
    'secure_link': '',
    # http://nginx.org/en/docs/http/ngx_http_secure_link_module.html#var_secure_link_expires
    'secure_link_expires': '',
    # http://nginx.org/en/docs/http/ngx_http_core_module.html#var_sent_http_
    'sent_http_': '',
    # http://nginx.org/en/docs/http/ngx_http_core_module.html#var_server_name
    'server_name': '',

    # "Secure" variables that can't content or strictly limited user input

    # http://nginx.org/en/docs/http/ngx_http_browser_module.html#var_ancient_browser
    'ancient_browser': None,
    # http://nginx.org/en/docs/http/ngx_http_core_module.html#var_binary_remote_addr
    # http://nginx.org/en/docs/stream/ngx_stream_core_module.html#var_binary_remote_addr
    'binary_remote_addr': None,
    # http://nginx.org/en/docs/http/ngx_http_core_module.html#var_body_bytes_sent
    'body_bytes_sent': None,
    # http://nginx.org/en/docs/stream/ngx_stream_core_module.html#var_bytes_received
    'bytes_received': None,
    # http://nginx.org/en/docs/http/ngx_http_core_module.html#var_bytes_sent
    # http://nginx.org/en/docs/http/ngx_http_log_module.html#var_bytes_sent
    # http://nginx.org/en/docs/stream/ngx_stream_core_module.html#var_bytes_sent
    'bytes_sent': None,
    # http://nginx.org/en/docs/http/ngx_http_core_module.html#var_connection
    # http://nginx.org/en/docs/http/ngx_http_log_module.html#var_connection
    # http://nginx.org/en/docs/stream/ngx_stream_core_module.html#var_connection
    'connection': None,
    # http://nginx.org/en/docs/http/ngx_http_core_module.html#var_connection_requests
    # http://nginx.org/en/docs/http/ngx_http_log_module.html#var_connection_requests
    'connection_requests': None,
    # http://nginx.org/en/docs/http/ngx_http_stub_status_module.html#var_connections_active
    'connections_active': None,
    # http://nginx.org/en/docs/http/ngx_http_stub_status_module.html#var_connections_reading
    'connections_reading': None,
    # http://nginx.org/en/docs/http/ngx_http_stub_status_module.html#var_connections_waiting
    'connections_waiting': None,
    # http://nginx.org/en/docs/http/ngx_http_stub_status_module.html#var_connections_writing
    'connections_writing': None,
    # http://nginx.org/en/docs/http/ngx_http_core_module.html#var_content_length
    'content_length': None,
    # http://nginx.org/en/docs/http/ngx_http_ssi_module.html#var_date_gmt
    'date_gmt': None,
    # http://nginx.org/en/docs/http/ngx_http_ssi_module.html#var_date_local
    'date_local': None,
    # http://nginx.org/en/docs/http/ngx_http_core_module.html#var_document_root
    'document_root': '/etc/nginx',
    # http://nginx.org/en/docs/http/ngx_http_geoip_module.html
    # http://nginx.org/en/docs/stream/ngx_stream_geoip_module.html
    'geoip_': None,
    # http://nginx.org/en/docs/http/ngx_http_gzip_module.html#var_gzip_ratio
    'gzip_ratio': None,
    # http://nginx.org/en/docs/http/ngx_http_v2_module.html#var_http2
    'http2': None,
    # http://nginx.org/en/docs/http/ngx_http_core_module.html#var_https
    'https': None,
    # http://nginx.org/en/docs/http/ngx_http_referer_module.html#var_invalid_referer
    'invalid_referer': None,
    # http://nginx.org/en/docs/http/ngx_http_core_module.html#var_is_args
    'is_args': None,
    # http://nginx.org/en/docs/http/ngx_http_auth_jwt_module.html
    'jwt_': None,
    # http://nginx.org/en/docs/http/ngx_http_browser_module.html#var_modern_browser
    'modern_browser': None,
    # http://nginx.org/en/docs/http/ngx_http_core_module.html#var_msec
    # http://nginx.org/en/docs/http/ngx_http_log_module.html#var_msec
    # http://nginx.org/en/docs/stream/ngx_stream_core_module.html#var_msec
    'msec': None,
    # http://nginx.org/en/docs/http/ngx_http_browser_module.html#var_msie
    'msie': None,
    # http://nginx.org/en/docs/http/ngx_http_core_module.html#var_nginx_version
    # http://nginx.org/en/docs/stream/ngx_stream_core_module.html#var_nginx_version
    'nginx_version': None,
    # http://nginx.org/en/docs/http/ngx_http_core_module.html#var_pid
    # http://nginx.org/en/docs/stream/ngx_stream_core_module.html#var_pid
    'pid': None,
    # http://nginx.org/en/docs/http/ngx_http_core_module.html#var_pipe
    # http://nginx.org/en/docs/http/ngx_http_log_module.html#var_pipe
    'pipe': None,
    # http://nginx.org/en/docs/stream/ngx_stream_core_module.html#var_protocol
    'protocol': None,
    # http://nginx.org/en/docs/http/ngx_http_realip_module.html#var_realip_remote_addr
    # http://nginx.org/en/docs/stream/ngx_stream_realip_module.html#var_realip_remote_addr
    # http://nginx.org/en/docs/http/ngx_http_realip_module.html#var_realip_remote_port
    # http://nginx.org/en/docs/stream/ngx_stream_realip_module.html#var_realip_remote_port
    'realip_remote_port': None,
    # http://nginx.org/en/docs/http/ngx_http_core_module.html#var_remote_addr
    # http://nginx.org/en/docs/stream/ngx_stream_core_module.html#var_remote_addr
    'remote_addr': None,
    # http://nginx.org/en/docs/http/ngx_http_core_module.html#var_remote_port
    # http://nginx.org/en/docs/stream/ngx_stream_core_module.html#var_remote_port
    'remote_port': None,
    # http://nginx.org/en/docs/http/ngx_http_core_module.html#var_request_body_file
    'request_body_file': None,
    # http://nginx.org/en/docs/http/ngx_http_core_module.html#var_request_length
    # http://nginx.org/en/docs/http/ngx_http_log_module.html#var_request_length
    'request_length': None,
    # http://nginx.org/en/docs/http/ngx_http_core_module.html#var_request_method
    'request_method': None,
    # http://nginx.org/en/docs/http/ngx_http_core_module.html#var_request_time
    # http://nginx.org/en/docs/http/ngx_http_log_module.html#var_request_time
    'request_time': None,
    # http://nginx.org/en/docs/http/ngx_http_core_module.html#var_scheme
    'scheme': None,
    # http://nginx.org/en/docs/http/ngx_http_core_module.html#var_server_addr
    # http://nginx.org/en/docs/stream/ngx_stream_core_module.html#var_server_addr
    'server_addr': None,
    # http://nginx.org/en/docs/http/ngx_http_core_module.html#var_server_port
    # http://nginx.org/en/docs/stream/ngx_stream_core_module.html#var_server_port
    'server_port': None,
    # http://nginx.org/en/docs/http/ngx_http_core_module.html#var_server_protocol
    'server_protocol': None,
    # http://nginx.org/en/docs/http/ngx_http_session_log_module.html#var_session_log_binary_id
    'session_log_binary_id': None,
    # http://nginx.org/en/docs/http/ngx_http_session_log_module.html#var_session_log_id
    'session_log_id': None,
    # http://nginx.org/en/docs/stream/ngx_stream_core_module.html#var_session_time
    'session_time': None,
    # http://nginx.org/en/docs/http/ngx_http_spdy_module.html#var_spdy
    'spdy': None,
    # http://nginx.org/en/docs/http/ngx_http_spdy_module.html#var_spdy_request_priority
    'spdy_request_priority': None,
    # http://nginx.org/en/docs/http/ngx_http_ssl_module.html
    # http://nginx.org/en/docs/stream/ngx_stream_ssl_module.html
    'ssl_': None,
    # http://nginx.org/en/docs/http/ngx_http_core_module.html#var_status
    # http://nginx.org/en/docs/http/ngx_http_log_module.html#var_status
    # http://nginx.org/en/docs/stream/ngx_stream_core_module.html#var_status
    'status': None,
    # http://nginx.org/en/docs/http/ngx_http_core_module.html
    'tcpinfo_': None,
    # http://nginx.org/en/docs/http/ngx_http_core_module.html
    # http://nginx.org/en/docs/http/ngx_http_log_module.html
    # http://nginx.org/en/docs/stream/ngx_stream_core_module.html
    'time_iso8601': None,
    # http://nginx.org/en/docs/http/ngx_http_core_module.html
    # http://nginx.org/en/docs/http/ngx_http_log_module.html
    # http://nginx.org/en/docs/stream/ngx_stream_core_module.html
    'time_local': None,
    # http://nginx.org/en/docs/http/ngx_http_userid_module.html#var_uid_got
    'uid_got': None,
    # http://nginx.org/en/docs/http/ngx_http_userid_module.html#var_uid_reset
    'uid_reset': None,
    # http://nginx.org/en/docs/http/ngx_http_userid_module.html#var_uid_set
    'uid_set': None,
    # http://nginx.org/en/docs/http/ngx_http_upstream_module.html#var_upstream_addr
    # http://nginx.org/en/docs/stream/ngx_stream_upstream_module.html#var_upstream_addr
    'upstream_addr': None,
    # http://nginx.org/en/docs/http/ngx_http_upstream_module.html#var_upstream_bytes_received
    # http://nginx.org/en/docs/stream/ngx_stream_upstream_module.html#var_upstream_bytes_received
    'upstream_bytes_received': None,
    # http://nginx.org/en/docs/stream/ngx_stream_upstream_module.html#var_upstream_bytes_sent
    'upstream_bytes_sent': None,
    # http://nginx.org/en/docs/http/ngx_http_upstream_module.html#var_upstream_cache_status
    'upstream_cache_status': None,
    # http://nginx.org/en/docs/http/ngx_http_upstream_module.html#var_upstream_connect_time
    # http://nginx.org/en/docs/stream/ngx_stream_upstream_module.html#var_upstream_connect_time
    'upstream_connect_time': None,
    # http://nginx.org/en/docs/stream/ngx_stream_upstream_module.html#var_upstream_first_byte_time
    'upstream_first_byte_time': None,
    # http://nginx.org/en/docs/http/ngx_http_upstream_module.html#var_upstream_header_time
    'upstream_header_time': None,
    # http://nginx.org/en/docs/http/ngx_http_upstream_module.html#var_upstream_response_length
    'upstream_response_length': None,
    # http://nginx.org/en/docs/http/ngx_http_upstream_module.html#var_upstream_response_time
    'upstream_response_time': None,
    # http://nginx.org/en/docs/stream/ngx_stream_upstream_module.html#var_upstream_session_time
    'upstream_session_time': None,
    # http://nginx.org/en/docs/http/ngx_http_upstream_module.html#var_upstream_status
    'upstream_status': None
}


def is_builtin(name):
    if isinstance(name, int):
        # Indexed variables can't be builtin
        return False
    for builtin in BUILTIN_VARIABLES:
        if builtin.endswith('_'):
            if name.startswith(builtin):
                return True
        elif name == builtin:
            return True
    return False


def builtin_var(name):
    for builtin, regexp in BUILTIN_VARIABLES.items():
        if builtin.endswith('_'):
            if not name.startswith(builtin):
                continue
        elif name != builtin:
            continue

        if regexp:
            return Variable(name=name, value=Regexp(regexp, strict=True, case_sensitive=False))
        return Variable(name=name, value='builtin', have_script=False)
    return None



import gixy
from gixy.core.plugins_manager import PluginsManager
from gixy.core.context import get_context, pop_context, push_context, purge_context
from gixy.parser.nginx_parser import NginxParser
from gixy.core.config import Config


class Manager(object):
    def __init__(self, config=None):
        self.root = None
        self.parser = None
        self.auditor = None
        self.config = config or Config()
        self.stats = {gixy.severity.UNSPECIFIED: 0,
                      gixy.severity.LOW: 0,
                      gixy.severity.MEDIUM: 0,
                      gixy.severity.HIGH: 0}

    def audit(self, file_path):
        self.auditor = PluginsManager(config=self.config)
        self.parser = NginxParser(file_path, allow_includes=self.config.allow_includes)
        self.root = self.parser.parse(file_path)
        push_context(self.root)
        self._audit_recursive(self.root.children)

    def get_results(self):
        for plugin in self.auditor.plugins:
            if plugin.issues:
                self.stats[plugin.severity] += len(plugin.issues)
                yield plugin

    def _audit_recursive(self, tree):
        for directive in tree:
            self._update_variables(directive)
            self.auditor.audit(directive)
            if directive.is_block:
                if directive.self_context:
                    push_context(directive)
                self._audit_recursive(directive.children)
                if directive.self_context:
                    pop_context()

    def _update_variables(self, directive):
        # TODO(buglloc): finish him!
        if not directive.provide_variables:
            return

        context = get_context()
        for var in directive.variables:
            if var.name == 0:
                # All regexps must clean indexed variables
                context.clear_index_vars()
            context.add_var(var.name, var)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        purge_context()



import os
from gixy.formatters.base import BaseFormatter

FORMATTERS = {}


def import_formatters():
    files_list = os.listdir(os.path.dirname(__file__))
    for formatter_file in files_list:
        if not formatter_file.endswith(".py") or formatter_file.startswith('_'):
            continue
        __import__('gixy.formatters.' + os.path.splitext(formatter_file)[0], None, None, [''])


def get_all():
    if len(FORMATTERS):
        return FORMATTERS

    import_formatters()
    for klass in BaseFormatter.__subclasses__():
        FORMATTERS[klass.__name__.replace('Formatter', '').lower()] = klass

    return FORMATTERS



import gixy
from gixy.plugins.plugin import Plugin
from gixy.core.variable import compile_script


class http_splitting(Plugin):
    """
    Insecure examples:
        rewrite ^ http://$host$uri;
        return 301 http://$host$uri;
        proxy_set_header "X-Original-Uri" $uri;
        proxy_pass http://upstream$document_uri;

        location ~ /proxy/(a|b)/(\W*)$ {
            set $path $2;
            proxy_pass http://storage/$path;
        }
    """

    summary = 'Possible HTTP-Splitting vulnerability.'
    severity = gixy.severity.HIGH
    description = 'Using variables that can contain "\\n" may lead to http injection.'
    help_url = 'https://github.com/yandex/gixy/blob/master/docs/ru/plugins/httpsplitting.md'
    directives = ['rewrite', 'return', 'add_header', 'proxy_set_header', 'proxy_pass']

    def audit(self, directive):
        value = _get_value(directive)
        if not value:
            return

        for var in compile_script(value):
            if not var.can_contain('\n'):
                continue
            reason = 'At least variable "${var}" can contain "\\n"'.format(var=var.name)
            self.add_issue(directive=[directive] + var.providers, reason=reason)


def _get_value(directive):
    if directive.name == 'proxy_pass' and len(directive.args) >= 1:
        return directive.args[0]
    elif len(directive.args) >= 2:
        return directive.args[1]
    return None



UNSPECIFIED = 'UNSPECIFIED'
LOW = 'LOW'
MEDIUM = 'MEDIUM'
HIGH = 'HIGH'
ALL = [UNSPECIFIED, LOW, MEDIUM, HIGH]


def is_acceptable(current_severity, min_severity):
    return ALL.index(current_severity) >= ALL.index(min_severity)



from __future__ import absolute_import

from gixy.directives import block


class BaseFormatter(object):
    skip_parents = {block.Root, block.HttpBlock}

    def format_reports(self, reports, stats):
        raise NotImplementedError("Formatter must override format_reports function")

    def format(self, manager):
        reports = []
        for result in manager.get_results():
            report = self._prepare_result(manager.root,
                                          summary=result.summary,
                                          severity=result.severity,
                                          description=result.description,
                                          issues=result.issues,
                                          plugin=result.name,
                                          help_url=result.help_url)
            reports.extend(report)

        return self.format_reports(reports, manager.stats)

    def _prepare_result(self, root, issues, severity, summary, description, plugin, help_url):
        result = {}
        for issue in issues:
            report = dict(
                plugin=plugin,
                summary=issue.summary or summary,
                severity=issue.severity or severity,
                description=issue.description or description,
                help_url=issue.help_url or help_url,
                reason=issue.reason or '',
            )
            key = ''.join(report.values())
            report['directives'] = issue.directives
            if key in result:
                result[key]['directives'].extend(report['directives'])
            else:
                result[key] = report

        for report in result.values():
            if report['directives']:
                config = self._resolve_config(root, report['directives'])
            else:
                config = ''

            del report['directives']
            report['config'] = config
            yield report

    def _resolve_config(self, root, directives):
        points = set()
        for directive in directives:
            points.add(directive)
            points.update(p for p in directive.parents)

        result = self._traverse_tree(root, points, 0)
        return '\n'.join(result)

    def _traverse_tree(self, tree, points, level):
        result = []
        for leap in tree.children:
            if leap not in points:
                continue
            printable = type(leap) not in self.skip_parents
            # Special hack for includes
            # TODO(buglloc): fix me
            have_parentheses = type(leap) != block.IncludeBlock

            if printable:
                if leap.is_block:
                    result.append('')
                directive = str(leap).replace('\n', '\n' + '\t' * (level + 1))
                result.append('{:s}{:s}'.format('\t' * level, directive))

            if leap.is_block:
                result.extend(self._traverse_tree(leap, points, level + 1 if printable else level))
                if printable and have_parentheses:
                    result.append('{:s}}}'.format('\t' * level))

        return result



from __future__ import absolute_import
from jinja2 import Environment, PackageLoader

from gixy.formatters.base import BaseFormatter


class ConsoleFormatter(BaseFormatter):
    def __init__(self):
        env = Environment(loader=PackageLoader('gixy', 'formatters/templates'), trim_blocks=True, lstrip_blocks=True)
        self.template = env.get_template('console.j2')

    def format_reports(self, reports, stats):
        return self.template.render(issues=reports, stats=stats)



from cached_property import cached_property

from gixy.directives.directive import Directive
from gixy.core.variable import Variable
from gixy.core.regexp import Regexp


def get_overrides():
    result = {}
    for klass in Block.__subclasses__():
        if not klass.nginx_name:
            continue

        if not klass.__name__.endswith('Block'):
            continue

        result[klass.nginx_name] = klass
    return result


class Block(Directive):
    nginx_name = None
    is_block = True
    self_context = True

    def __init__(self, name, args):
        super(Block, self).__init__(name, args)
        self.children = []

    def some(self, name, flat=True):
        for child in self.children:
            if child.name == name:
                return child
            if flat and child.is_block and not child.self_context:
                result = child.some(name, flat=flat)
                if result:
                    return result
        return None

    def find(self, name, flat=False):
        result = []
        for child in self.children:
            if child.name == name:
                result.append(child)
            if flat and child.is_block and not child.self_context:
                result += child.find(name)
        return result

    def find_recursive(self, name):
        result = []
        for child in self.children:
            if child.name == name:
                result.append(child)
            if child.is_block:
                result += child.find_recursive(name)
        return result

    def append(self, directive):
        directive.set_parent(self)
        self.children.append(directive)

    def __str__(self):
        return '{} {} {}'.format(self.name, ' '.join(self.args), '{')


class Root(Block):
    nginx_name = None

    def __init__(self):
        super(Root, self).__init__(None, [])


class HttpBlock(Block):
    nginx_name = 'http'

    def __init__(self, name, args):
        super(HttpBlock, self).__init__(name, args)


class ServerBlock(Block):
    nginx_name = 'server'

    def __init__(self, name, args):
        super(ServerBlock, self).__init__(name, args)

    def get_names(self):
        return self.find('server_name')

    def __str__(self):
        server_names = [str(sn) for sn in self.find('server_name')]
        if server_names:
            return 'server {{\n{}'.format('\n'.join(server_names[:2]))
        return 'server {'


class LocationBlock(Block):
    nginx_name = 'location'
    provide_variables = True

    def __init__(self, name, args):
        super(LocationBlock, self).__init__(name, args)
        if len(args) == 2:
            self.modifier, self.path = args
        else:
            self.modifier = None
            self.path = args[0]

    @property
    def is_internal(self):
        return self.some('internal') is not None

    @cached_property
    def variables(self):
        if not self.modifier or self.modifier not in ('~', '~*'):
            return []

        regexp = Regexp(self.path, case_sensitive=self.modifier == '~')
        result = []
        for name, group in regexp.groups.items():
            result.append(Variable(name=name, value=group, boundary=None, provider=self))
        return result


class IfBlock(Block):
    nginx_name = 'if'
    self_context = False

    def __init__(self, name, args):
        super(IfBlock, self).__init__(name, args)
        self.operand = None
        self.value = None
        self.variable = None

        if len(args) == 1:
            # if ($slow)
            self.variable = args[0]
        elif len(args) == 2:
            # if (!-e $foo)
            self.operand, self.value = args
        elif len(args) == 3:
            # if ($request_method = POST)
            self.variable, self.operand, self.value = args
        else:
            raise Exception('Unknown "if" definition')

    def __str__(self):
        return '{} ({}) {{'.format(self.name, ' '.join(self.args))


class IncludeBlock(Block):
    nginx_name = 'include'
    self_context = False

    def __init__(self, name, args):
        super(IncludeBlock, self).__init__(name, args)
        self.file_path = args[0]

    def __str__(self):
        return 'include {};'.format(self.file_path)


class MapBlock(Block):
    nginx_name = 'map'
    self_context = False
    provide_variables = True

    def __init__(self, name, args):
        super(MapBlock, self).__init__(name, args)
        self.source = args[0]
        self.variable = args[1].strip('$')

    @cached_property
    def variables(self):
        # TODO(buglloc): Finish him!
        return [Variable(name=self.variable, value='', boundary=None, provider=self, have_script=False)]



from __future__ import absolute_import

import json

from gixy.formatters.base import BaseFormatter


class JsonFormatter(BaseFormatter):
    def format_reports(self, reports, stats):
        return json.dumps(reports, sort_keys=True, indent=2, separators=(',', ': '))



import re
from setuptools import setup, find_packages

with open('gixy/__init__.py', 'r') as fd:
    version = re.search(r'^version\s*=\s*[\'"]([^\'"]*)[\'"]',
                        fd.read(), re.MULTILINE).group(1)

if not version:
    raise RuntimeError('Cannot find version information')

setup(
    name='gixy',
    version=version,
    description='Nginx configuration [sec]analyzer',
    keywords='nginx security lint static-analysis',
    author='Yandex IS Team',
    author_email='buglloc@yandex.ru',
    url='https://github.com/yandex/gixy',
    install_requires=[
        'pyparsing>=1.5.5',
        'cached-property>=1.2.0',
        'argparse>=1.4.0',
        'six>=1.1.0',
        'Jinja2>=2.8',
        'ConfigArgParse>=0.11.0'
    ],
    entry_points={
        'console_scripts': ['gixy=gixy.cli.main:main'],
    },
    packages=find_packages(exclude=['tests', 'tests.*']),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: System Administrators',
        'Intended Audience :: Developers',
        'Topic :: Security',
        'Topic :: Software Development :: Quality Assurance',
        'Topic :: Software Development :: Testing'
    ],
    include_package_data=True,
    )



import gixy
from gixy.plugins.plugin import Plugin


class add_header_multiline(Plugin):
    """
    Insecure example:
add_header Content-Security-Policy "
    default-src: 'none';
    img-src data: https://mc.yandex.ru https://yastatic.net *.yandex.net https://mc.yandex.${tld} https://mc.yandex.ru;
    font-src data: https://yastatic.net;";
    """
    summary = 'Found a multi-line header.'
    severity = gixy.severity.LOW
    description = ('Multi-line headers are deprecated (see RFC 7230). '
                   'Some clients never supports them (e.g. IE/Edge).')
    help_url = 'https://github.com/yandex/gixy/blob/master/docs/ru/plugins/addheadermultiline.md'
    directives = ['add_header', 'more_set_headers']

    def audit(self, directive):
        header_values = get_header_values(directive)
        for value in header_values:
            if '\n\x20' in value or '\n\t' in value:
                self.add_issue(directive=directive)
                break


def get_header_values(directive):
    if directive.name == 'add_header':
        return [directive.args[1]]

    # See headers more documentation: https://github.com/openresty/headers-more-nginx-module#description
    result = []
    skip_next = False
    for arg in directive.args:
        if arg in {'-s', '-t'}:
            # Skip next value, because it's not a header
            skip_next = True
        elif arg.startswith('-'):
            # Skip any options
            pass
        elif skip_next:
            skip_next = False
        elif not skip_next:
            result.append(arg)
    return result



"""NginxParser is a member object of the NginxConfigurator class."""
import os
import glob
import logging
import fnmatch

from pyparsing import ParseException

from gixy.parser import raw_parser
from gixy.directives import block, directive

LOG = logging.getLogger(__name__)


class NginxParser(object):

    def __init__(self, file_path, allow_includes=True):
        self.base_file_path = file_path
        self.cwd = os.path.dirname(file_path)
        self.configs = {}
        self.is_dump = False
        self.allow_includes = allow_includes
        self.directives = {}
        self._init_directives()

    def parse(self, file_path, root=None):
        LOG.debug("Parse file: {}".format(file_path))

        if not root:
            root = block.Root()

        try:
            parser = raw_parser.RawParser()
            parsed = parser.parse(file_path)
        except ParseException as e:
            LOG.error('Failed to parse config "{file}": {error}'.format(file=file_path, error=str(e)))
            return root

        if len(parsed) and parsed[0].getName() == 'file_delimiter':
            #  Were parse nginx dump
            LOG.info('Switched to parse nginx configuration dump.')
            root_filename = self._prepare_dump(parsed)
            self.is_dump = True
            self.cwd = os.path.dirname(root_filename)
            parsed = self.configs[root_filename]

        self.parse_block(parsed, root)
        return root

    def parse_block(self, parsed_block, parent):
        for parsed in parsed_block:
            parsed_type = parsed.getName()
            parsed_name = parsed[0]
            parsed_args = parsed[1:]
            if parsed_type == 'include':
                # TODO: WTF?!
                self._resolve_include(parsed_args, parent)
            else:
                directive_inst = self.directive_factory(parsed_type, parsed_name, parsed_args)
                if directive_inst:
                    parent.append(directive_inst)

    def directive_factory(self, parsed_type, parsed_name, parsed_args):
        klass = self._get_directive_class(parsed_type, parsed_name)
        if not klass:
            return None

        if klass.is_block:
            args = [str(v).strip() for v in parsed_args[0]]
            children = parsed_args[1]

            inst = klass(parsed_name, args)
            self.parse_block(children, inst)
            return inst
        else:
            args = [str(v).strip() for v in parsed_args]
            return klass(parsed_name, args)

    def _get_directive_class(self, parsed_type, parsed_name):
        if parsed_type in self.directives and parsed_name in self.directives[parsed_type]:
            return self.directives[parsed_type][parsed_name]
        elif parsed_type == 'block':
            return block.Block
        elif parsed_type == 'directive':
            return directive.Directive
        elif parsed_type == 'unparsed_block':
            LOG.warning('Skip unparseable block: "%s"', parsed_name)
            return None
        else:
            return None

    def _init_directives(self):
        self.directives['block'] = block.get_overrides()
        self.directives['directive'] = directive.get_overrides()

    def _resolve_include(self, args, parent):
        pattern = args[0]
        #  TODO(buglloc): maybe file providers?
        if self.is_dump:
            return self._resolve_dump_include(pattern=pattern, parent=parent)
        if not self.allow_includes:
            LOG.debug('Includes are disallowed, skip: {}'.format(pattern))
            return

        return self._resolve_file_include(pattern=pattern, parent=parent)

    def _resolve_file_include(self, pattern, parent):
        path = os.path.join(self.cwd, pattern)
        file_path = None
        for file_path in glob.iglob(path):
            include = block.IncludeBlock('include', [file_path])
            parent.append(include)
            self.parse(file_path, include)

        if not file_path:
            LOG.warning("File not found: {}".format(path))

    def _resolve_dump_include(self, pattern, parent):
        path = os.path.join(self.cwd, pattern)
        founded = False
        for file_path, parsed in self.configs.items():
            if fnmatch.fnmatch(file_path, path):
                founded = True
                include = block.IncludeBlock('include', [file_path])
                parent.append(include)
                self.parse_block(parsed, include)

        if not founded:
            LOG.warning("File not found: {}".format(path))

    def _prepare_dump(self, parsed_block):
        filename = ''
        root_filename = ''
        for parsed in parsed_block:
            if parsed.getName() == 'file_delimiter':
                if not filename:
                    root_filename = parsed[0]
                filename = parsed[0]
                self.configs[filename] = []
                continue
            self.configs[filename].append(parsed)
        return root_filename



import re

import gixy
from gixy.plugins.plugin import Plugin
from gixy.core.context import get_context
from gixy.core.variable import compile_script


class ssrf(Plugin):
    """
    Insecure examples:
        location ~ /proxy/(.*)/(.*)/(.*)$ {
            set $scheme $1;
            set $host $2;
            set $path $3;
            proxy_pass $scheme://$host/$path;
        }

        location /proxy/ {
            proxy_pass $arg_some;
        }
    """

    summary = 'Possible SSRF (Server Side Request Forgery) vulnerability.'
    severity = gixy.severity.HIGH
    description = 'The configuration may allow attacker to create a arbitrary requests from the vulnerable server.'
    help_url = 'https://github.com/yandex/gixy/blob/master/docs/ru/plugins/ssrf.md'
    directives = ['proxy_pass']

    def __init__(self, config):
        super(ssrf, self).__init__(config)
        self.parse_uri_re = re.compile(r'(?P<scheme>[^?#/)]+://)?(?P<host>[^?#/)]+)')

    def audit(self, directive):
        value = directive.args[0]
        if not value:
            return

        context = get_context()
        if context.block.name == 'location' and context.block.is_internal:
            # Exclude internal locations
            return

        parsed = self.parse_uri_re.match(value)
        if not parsed:
            return

        res = self._check_script(parsed.group('scheme'), directive)
        if not res:
            self._check_script(parsed.group('host'), directive)

    def _check_script(self, script, directive):
        for var in compile_script(script):
            if var.must_contain('/'):
                # Skip variable checks
                return False
            if var.can_contain('.'):
                # Yay! Our variable can contain any symbols!
                reason = 'At least variable "${var}" can contain untrusted user input'.format(var=var.name)
                self.add_issue(directive=[directive] + var.providers, reason=reason)
                return True
        return False
